/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *  
 *    http://www.apache.org/licenses/LICENSE-2.0
 *  
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License. 
 *  
 */
package org.apache.directory.server.core.partition.impl.btree.jdbm;


import org.apache.directory.server.core.DirectoryService;
import org.apache.directory.server.core.entry.ServerEntry;
import org.apache.directory.server.core.interceptor.context.AddOperationContext;
import org.apache.directory.server.core.interceptor.context.BindOperationContext;
import org.apache.directory.server.core.interceptor.context.ModifyOperationContext;
import org.apache.directory.server.core.interceptor.context.MoveAndRenameOperationContext;
import org.apache.directory.server.core.interceptor.context.MoveOperationContext;
import org.apache.directory.server.core.interceptor.context.RenameOperationContext;
import org.apache.directory.server.core.interceptor.context.UnbindOperationContext;
import org.apache.directory.server.core.partition.Oid;
import org.apache.directory.server.core.partition.Partition;
import org.apache.directory.server.core.partition.impl.btree.BTreePartition;
import org.apache.directory.server.core.partition.impl.btree.DefaultOptimizer;
import org.apache.directory.server.core.partition.impl.btree.DefaultSearchEngine;
import org.apache.directory.server.core.partition.impl.btree.ExpressionEnumerator;
import org.apache.directory.server.core.partition.impl.btree.ExpressionEvaluator;
import org.apache.directory.server.core.partition.impl.btree.Index;
import org.apache.directory.server.core.partition.impl.btree.IndexNotFoundException;
import org.apache.directory.server.core.partition.impl.btree.NoOpOptimizer;
import org.apache.directory.server.schema.registries.Registries;
import org.apache.directory.shared.ldap.exception.LdapAuthenticationNotSupportedException;
import org.apache.directory.shared.ldap.message.ResultCodeEnum;
import org.apache.directory.shared.ldap.name.LdapDN;

import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import java.io.File;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;


/**
 * A {@link Partition} that stores entries in
 * <a href="http://jdbm.sourceforge.net/">JDBM</a> database.
 *
 * @org.apache.xbean.XBean
 * @author <a href="mailto:dev@directory.apache.org">Apache Directory Project</a>
 * @version $Rev: 652410 $
 */
public class JdbmPartition extends BTreePartition
{
    private JdbmStore store;
    private boolean optimizerEnabled = true;
    private Set<Index> indexedAttributes;

    
    // ------------------------------------------------------------------------
    // C O N S T R U C T O R S
    // ------------------------------------------------------------------------


    /**
     * Creates a store based on JDBM B+Trees.
     */
    public JdbmPartition()
    {
        store = new JdbmStore();
        indexedAttributes = new HashSet<Index>();
    }


    // ------------------------------------------------------------------------
    // C O N F I G U R A T I O N   M E T H O D S
    // ------------------------------------------------------------------------


    public String getSuffix()
    {
        return super.suffix;
    }


    public void setSuffix( String suffix )
    {
        super.suffix = suffix;
    }


    public void setIndexedAttributes( Set<Index> indexedAttributes )
    {
        this.indexedAttributes = indexedAttributes;
    }


    public Set<Index> getIndexedAttributes()
    {
        return indexedAttributes;
    }


    public boolean isOptimizerEnabled()
    {
        return optimizerEnabled;
    }


    public void setOptimizerEnabled( boolean optimizerEnabled )
    {
        this.optimizerEnabled = optimizerEnabled;
    }


    public void setSyncOnWrite( boolean syncOnWrite )
    {
        store.setSyncOnWrite( syncOnWrite );
    }


    public boolean isSyncOnWrite()
    {
        return store.isSyncOnWrite();
    }


    // ------------------------------------------------------------------------
    // E N D   C O N F I G U R A T I O N   M E T H O D S
    // ------------------------------------------------------------------------


    public void setRegistries( Registries registries )
    {
        initRegistries( registries );
    }


    protected void initRegistries( Registries registries )
    {
        this.registries = registries;
        ExpressionEvaluator evaluator = new ExpressionEvaluator( this, registries );
        ExpressionEnumerator enumerator = new ExpressionEnumerator( this, registries.getAttributeTypeRegistry(), evaluator );
        this.searchEngine = new DefaultSearchEngine( this, evaluator, enumerator, optimizer );
        store.initRegistries( registries );
    }


    public final void init( DirectoryService directoryService ) throws NamingException
    {
        // setup optimizer and registries for parent
        if ( ! optimizerEnabled )
        {
            optimizer = new NoOpOptimizer();
        }
        else
        {
            optimizer = new DefaultOptimizer( this );
        }

        initRegistries( directoryService.getRegistries() );
        
        // initialize the store
        store.setCacheSize( cacheSize );
        store.setContextEntry( contextEntry );
        store.setName( id );
        store.setSuffixDn( suffix );
        store.setWorkingDirectory( new File( directoryService.getWorkingDirectory().getPath() + File.separator + id ) );

        Set<JdbmIndex> userIndices = new HashSet<JdbmIndex>();
        for ( Index obj : indexedAttributes )
        {
            JdbmIndex index;

            if ( obj instanceof JdbmIndex )
            {
                index = ( JdbmIndex ) obj;
            }
            else
            {
                index = new JdbmIndex();
                index.setAttributeId( obj.getAttributeId() );
                index.setCacheSize( obj.getCacheSize() );
                index.setWkDirPath( obj.getWkDirPath() );
            }

            String oid = registries.getOidRegistry().getOid( index.getAttributeId() );
            if ( SYS_INDEX_OIDS.contains( registries.getOidRegistry().getOid( index.getAttributeId() ) ) )
            {
                if ( oid.equals( Oid.ALIAS ) )
                {
                    store.setAliasIndex( index );
                }
                else if ( oid.equals( Oid.EXISTANCE ) )
                {
                    store.setExistanceIndex( index );
                }
                else if ( oid.equals( Oid.HIERARCHY ) )
                {
                    store.setHierarchyIndex( index );
                }
                else if ( oid.equals( Oid.NDN ) )
                {
                    store.setNdnIndex( index );
                }
                else if ( oid.equals( Oid.ONEALIAS ) )
                {
                    store.setOneAliasIndex( index );
                }
                else if ( oid.equals( Oid.SUBALIAS ) )
                {
                    store.setSubAliasIndex( index );
                }
                else if ( oid.equals( Oid.UPDN ) )
                {
                    store.setUpdnIndex( index );
                }
                else
                {
                    throw new IllegalStateException( "Unrecognized system index " + oid );
                }
            }
            else
            {
                userIndices.add( index );
            }
            store.setUserIndices( userIndices );
            store.setEnableOptimizer( isOptimizerEnabled() );
        }

        store.init( registries );
    }


    public final void destroy()
    {
        store.destroy();
    }


    public final boolean isInitialized()
    {
        return store.isInitialized();
    }


    public final void sync() throws NamingException
    {
        store.sync();
    }


    // ------------------------------------------------------------------------
    // I N D E X   M E T H O D S
    // ------------------------------------------------------------------------


    public final void addIndexOn( Index index ) throws NamingException
    {
        if ( index instanceof JdbmIndex )
        {
            store.addIndex( ( JdbmIndex ) index );
        }
    }


    public final Index getExistanceIndex()
    {
        return store.getExistanceIndex();
    }


    public final void setExistanceIndexOn( Index index ) throws NamingException
    {
        if ( index instanceof JdbmIndex )
        {
            store.setExistanceIndex( ( JdbmIndex ) index );
        }
    }


    public final Index getHierarchyIndex()
    {
        return store.getHierarchyIndex();
    }


    public final void setHierarchyIndexOn( Index index ) throws NamingException
    {
        if ( index instanceof JdbmIndex )
        {
            store.setHierarchyIndex( ( JdbmIndex ) index );
        }
    }


    public final Index getAliasIndex()
    {
        return store.getAliasIndex();
    }


    public final void setAliasIndexOn( Index index ) throws NamingException
    {
        if ( index instanceof JdbmIndex )
        {
            store.setAliasIndex( ( JdbmIndex ) index );
        }
    }


    public final Index getOneAliasIndex()
    {
        return store.getOneAliasIndex();
    }


    public final void setOneAliasIndexOn( Index index ) throws NamingException
    {
        if ( index instanceof JdbmIndex )
        {
            store.setOneAliasIndex( ( JdbmIndex ) index );
        }
    }


    public final Index getSubAliasIndex()
    {
        return store.getSubAliasIndex();
    }


    public final void setSubAliasIndexOn( Index index ) throws NamingException
    {
        if ( index instanceof JdbmIndex )
        {
            store.setSubAliasIndex( ( JdbmIndex ) index );
        }
    }


    public final Index getUpdnIndex()
    {
        return store.getUpdnIndex();
    }


    public final void setUpdnIndexOn( Index index ) throws NamingException
    {
        if ( index instanceof JdbmIndex )
        {
            store.setUpdnIndex( ( JdbmIndex ) index );
        }
    }


    public final Index getNdnIndex()
    {
        return store.getNdnIndex();
    }


    public final void setNdnIndexOn( Index index ) throws NamingException
    {
        if ( index instanceof JdbmIndex )
        {
            store.setNdnIndex( ( JdbmIndex ) index );
        }
    }


    public final Iterator<String> getUserIndices()
    {
        return store.userIndices();
    }


    public final Iterator<String> getSystemIndices()
    {
        return store.systemIndices();
    }


    public final boolean hasUserIndexOn( String id ) throws NamingException
    {
        return store.hasUserIndexOn( id );
    }


    public final boolean hasSystemIndexOn( String id ) throws NamingException
    {
        return store.hasSystemIndexOn( id );
    }


    /**
     * @see org.apache.directory.server.core.partition.impl.btree.BTreePartition#getUserIndex(String)
     */
    public final Index getUserIndex( String id ) throws IndexNotFoundException
    {
        return store.getUserIndex( id );
    }


    /**
     * @see BTreePartition#getEntryId(String)
     */
    public final Index getSystemIndex( String id ) throws IndexNotFoundException
    {
        return store.getSystemIndex( id );
    }


    public final Long getEntryId( String dn ) throws NamingException
    {
        return store.getEntryId( dn );
    }


    public final String getEntryDn( Long id ) throws NamingException
    {
        return store.getEntryDn( id );
    }


    public final Long getParentId( String dn ) throws NamingException
    {
        return store.getParentId( dn );
    }


    public final Long getParentId( Long childId ) throws NamingException
    {
        return store.getParentId( childId );
    }


    public final String getEntryUpdn( Long id ) throws NamingException
    {
        return store.getEntryUpdn( id );
    }


    public final String getEntryUpdn( String dn ) throws NamingException
    {
        return store.getEntryUpdn( dn );
    }


    public final int count() throws NamingException
    {
        return store.count();
    }

    
    public final void add( AddOperationContext addContext ) throws NamingException
    {
        store.add( addContext.getDn(), addContext.getEntry() );
    }


    public final ServerEntry lookup( Long id ) throws NamingException
    {
        return store.lookup( id );
    }


    public final void delete( Long id ) throws NamingException
    {
        store.delete( id );
    }


    public final NamingEnumeration list( Long id ) throws NamingException
    {
        return store.list( id );
    }


    public final int getChildCount( Long id ) throws NamingException
    {
        return store.getChildCount( id );
    }


    public final LdapDN getSuffixDn()
    {
        return store.getSuffix();
    }

    public final LdapDN getUpSuffixDn()
    {
        return store.getUpSuffix();
    }


    public final ServerEntry getSuffixEntry() throws NamingException
    {
        return store.getSuffixEntry();
    }


    public final void setProperty( String propertyName, String propertyValue ) throws NamingException
    {
        store.setProperty( propertyName, propertyValue );
    }


    public final String getProperty( String propertyName ) throws NamingException
    {
        return store.getProperty( propertyName );
    }


    public final ServerEntry getIndices( Long id ) throws NamingException
    {
        return store.getIndices( id );
    }

    
    public final void modify( ModifyOperationContext modifyContext ) throws NamingException
    {
        store.modify( modifyContext.getDn(), modifyContext.getModItems() );
    }

    public final void rename( RenameOperationContext renameContext ) throws NamingException
    {
        store.rename( renameContext.getDn(), renameContext.getNewRdn(), renameContext.getDelOldDn() );
    }


    public final void moveAndRename( MoveAndRenameOperationContext moveAndRenameContext ) throws NamingException
    {
        store.move( moveAndRenameContext.getDn(), 
            moveAndRenameContext.getParent(), 
            moveAndRenameContext.getNewRdn(), 
            moveAndRenameContext.getDelOldDn() );
    }


    public final void move( MoveOperationContext moveContext ) throws NamingException
    {
        store.move( moveContext.getDn(), moveContext.getParent() );
    }


    public final void bind( LdapDN bindDn, byte[] credentials, List<String> mechanisms, String saslAuthId ) throws NamingException
    {
        if ( bindDn == null || credentials == null || mechanisms == null ||  saslAuthId == null )
        {
            // do nothing just using variables to prevent yellow lights : bad :)
        }
        
        // does nothing
        throw new LdapAuthenticationNotSupportedException(
                "Bind requests only tunnel down into partitions if there are no authenticators to handle the mechanism.\n"
                        + "Check to see if you have correctly configured authenticators for the server.",
                ResultCodeEnum.AUTH_METHOD_NOT_SUPPORTED );
    }
    

    public final void bind( BindOperationContext bindContext ) throws NamingException
    {
        // does nothing
        throw new LdapAuthenticationNotSupportedException(
            "Bind requests only tunnel down into partitions if there are no authenticators to handle the mechanism.\n"
                + "Check to see if you have correctly configured authenticators for the server.",
            ResultCodeEnum.AUTH_METHOD_NOT_SUPPORTED );
    }


    public final void unbind( UnbindOperationContext unbindContext ) throws NamingException
    {
    }
}
